#
# EDK BSP board generation for U-boot supporting Microblaze and PPC
#
# (C) Copyright 2007 Xilinx, Inc.
# Based on original code:
# (C) Copyright 2007 Michal Simek
#
# Michal SIMEK <monstr@monstr.eu>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307 USA
#
# ChangeLOG:
# 4.00.c Generating Kconfig.auto for Linux kernel with FDT support
# 4.00.b Separate U-BOOT & FDT part, Adding generation files for standalone,
#	generation adress of cache for Microblaze
# 4.00.a Redesign from Xilinx
# 2.00.b Remove problem with storing file
# 2.00.a Testing interrupt connection between Microblaze & INTC
# 1.01.c add simple FDT support - testing version
# 1.01.b add Microblaze HW_VER support, FLAGS, Uart16550. Remove MSS drivers
#	dependence, Timer handle throught INTC, BARREL shifter
# 1.01.a add MSR support and mch_opb_ddr
# 1.01.f add FSL support
# 1.01.e add IIC support
# 1.00.d clear Czech comments
# 1.00.c synchronization FLASH_MEMORY

# Globals variable
set version "U-BOOT v4.00.c"
set cpunumber 0
set periphery_array ""
set kconfig_file {}

proc uboot_drc {os_handle} {
	puts "\#--------------------------------------"
	puts "\# uboot BSP DRC...!"
	puts "\#--------------------------------------"
}

proc generate {os_handle} {
	generate_uboot $os_handle
}

# procedure post_generate
# This generates the drivers directory for uboot
# and runs the ltypes script

proc post_generate {lib_handle} {
}

proc generate_uboot {os_handle} {
	# ******************************************************************************
	# U-BOOT part
	# ******************************************************************************

	puts "\#--------------------------------------"
	puts "\# uboot BSP generate..."
	puts "\#--------------------------------------"

	# Open files and print GPL licence
	set config_file2 [open "config.mk" w]
	headerm $config_file2
	set config_file [open "xparameters.h" w]
	headerc $config_file

	global kconfig_file
	set kconfig_file [open "KconfigFDT.auto" w]
	headerm $kconfig_file

	# ******************************************************************************
	# print system clock
	set proc_handle [xget_libgen_proc_handle]
	set sys_clock [xget_sw_parameter_value $proc_handle "CORE_CLOCK_FREQ_HZ"]
	puts $config_file "/* System Clock Frequency */"
	puts $config_file "#define XILINX_CLOCK_FREQ\t$sys_clock\n"

	# Microblaze
	set hwproc_handle [xget_handle $proc_handle "IPINST"]
	set args [xget_hw_parameter_handle $hwproc_handle "*"]
	set proctype [xget_value $hwproc_handle "OPTION" "IPNAME"]
	switch $proctype {
		"microblaze" {
			# write only name of instance
			puts $config_file "/* Microblaze is [xget_hw_parameter_value $hwproc_handle "INSTANCE"] */"
			generate_comment "Definitions for MICROBLAZE0"

			foreach arg $args {
				set arg_name [xget_value $arg "NAME"]
				set arg_name [string map -nocase {C_ ""} $arg_name]
				set arg_value [xget_value $arg "VALUE"]
				#	puts $config_file "DEBUG $arg_name $arg_value"
				switch $arg_name {
					USE_MSR_INSTR {
						puts $config_file "#define XILINX_USE_MSR_INSTR\t$arg_value"
						generate_int_param "XILINX_MICROBLAZE0_USE_MSR_INSTR" $arg_value "USE_MSR_INSTR"
					}
					FSL_LINKS {
						puts $config_file "#define XILINX_FSL_NUMBER\t$arg_value"
					}
					HW_VER {
						puts $config_file2 "PLATFORM_CPPFLAGS += -mcpu=v$arg_value"
						generate_string_param "XILINX_MICROBLAZE0_HW_VER" $arg_value "Core version number"
					}
					USE_BARREL {
						if {[string match $arg_value "1"]} {
							puts $config_file2 "PLATFORM_CPPFLAGS += -mxl-barrel-shift"
						}
						generate_int_param "XILINX_MICROBLAZE0_USE_BARREL" $arg_value "USE_BARREL range (0:1)"
					}
					USE_DIV {
						if {[string match $arg_value "1"]} {
							puts $config_file2 "PLATFORM_CPPFLAGS += -mno-xl-soft-div"
						}
						#FIXME What is -mno-xl-hard-div
						generate_int_param "XILINX_MICROBLAZE0_USE_DIV" $arg_value "USE_DIV range (0:1)"
					}
					USE_HW_MUL {
						if {[string match $arg_value "1"]} {
							puts $config_file2 "PLATFORM_CPPFLAGS += -mno-xl-soft-mul"
						} else {
							puts $config_file2 "PLATFORM_CPPFLAGS += -mxl-soft-mul"
						}
						#FIXME What is -mxl-multiply-high???
						generate_int_param "XILINX_MICROBLAZE0_USE_HW_MULT" $arg_value "USE_HW_MULT range (0:1)"
					}
					USE_PCMP_INSTR {
						if {[string match $arg_value "1"]} {
							puts $config_file2 "PLATFORM_CPPFLAGS += -mxl-pattern-compare"
						}
#						generate_int_param "XILINX_MICROBLAZE0_USE_PCMP_INSTR" $arg_value "USE_PCMP_INSTR"
					}
					USE_FPU {
						if {[string match $arg_value "1"]} {
							puts $config_file2 "PLATFORM_CPPFLAGS += -mhard-float"
						}
						generate_int_param "XILINX_MICROBLAZE0_USE_FPU" $arg_value "USE_FPU range (0:1)"
					}
					PVR {
						puts $config_file "#define XILINX_PVR\t$arg_value"
					}
					FAMILY {
						generate_string_param "XILINX_MICROBLAZE0_FAMILY" $arg_value "Targetted FPGA family"
					}
					default {}
				}
			}

			# OPB bus resolve
			set dopb2 [xget_handle $hwproc_handle "BUS_INTERFACE" "DOPB"]
			set dopb [xget_value $dopb2 "VALUE"]
			set iopb [xget_handle $hwproc_handle "BUS_INTERFACE" "IOPB"]
			set iopb [xget_value $iopb "VALUE"]
			if { $dopb == $iopb } {
				set system_bus "$dopb"
				puts "System bus for instruction and data $dopb"
				#testing
				#	set bus [xget_sw_parameter_value $os_handle "opb_v20"]
				##	set bus_handle [xget_sw_ipinst_handle_from_processor $proc_handle $bus]
				##	set hodn [xget_sw_parameter_value $bus_handle "C_EXT_RESET_HIGH"]
				#	puts "fdf $bus fds"
				#	set clk [xget_handle $dopb "PORT" "OPB_Clk"]
				#	puts "$clk"
				#	set clk [xget_value $clk "VALUE"]
				#	error "$clk"
				#end testing
			} else {
				error "different microblaze architecture - dual busses $iopb $dopb"
			}
			puts $config_file ""
			uboot_intc $os_handle $proc_handle $config_file $config_file2 $system_bus
		}
		"ppc405" -
		"ppc405_virtex4" {
			puts "unsupported processor type $proctype\n"
#			error "ERROR $proctype not supported by U-BOOT yet"
		}
		default {
			error "This type of CPU is not supported by U-BOOT yet"
		}
	}
	close $kconfig_file
	close $config_file
	close $config_file2
}

#function for handling adress
proc uboot_addr_hex {handle name} {
	return [format "0x%08x" [uboot_value $handle $name]]
}

proc uboot_value {handle name} {
	set addr [xget_sw_parameter_value "$handle" "$name"]
	if {![llength $addr]} {
		error "Request for undefined value [xget_hw_name $handle]:$name"
	}
	return $addr
}


proc uboot_intc {os_handle proc_handle config_file config_file2 system_bus} {
# ******************************************************************************
# Interrupt controler
	set intc_handle [get_handle_to_intc $proc_handle "Interrupt"]
	if {[string match "" $intc_handle] || [string match -nocase "none" $intc_handle]} {
		puts $config_file "/* Interrupt controller not defined */"
	} else {
		puts $config_file "/* Interrupt controller is [xget_hw_name $intc_handle] */"
#FIXME redesign test_buses - give name of buses from IP
		test_buses $system_bus $intc_handle "SOPB"
#		Interrupt controller address
		puts $config_file "#define XILINX_INTC_BASEADDR\t[uboot_addr_hex $intc_handle "C_BASEADDR"]"

		set intc_value [uboot_value $intc_handle "C_NUM_INTR_INPUTS"]
		puts $config_file "#define XILINX_INTC_NUM_INTR_INPUTS\t$intc_value"
		set intc_value [expr $intc_value - 1]

		set port_list [xget_port_by_subtype $intc_handle "*"]
		foreach port $port_list {
			set name [xget_value $port "NAME"]
			if {[string match -nocase $name "intr"]} {
				set intc_irq [xget_value $port "VALUE"]
#DEBUG				puts $config_file "/* pripojene interrupty $name=$intc_irq */"
				set intc_signals [split $intc_irq "&"]
#				split the signals
#				DEBUG	puts $config_file "$signals"
			}
		}
		puts $config_file ""
# ****************************************************************************
# Timer part
# handle timer if exists intc
		set timer [xget_sw_parameter_value $os_handle "timer"]
		if {[string match "" $timer] || [string match -nocase "none" $timer]} {
			puts $config_file "/* Timer not defined */"
		} else {
			set timer_handle [xget_sw_ipinst_handle_from_processor $proc_handle $timer]
			#test for correct system bus
			test_buses $system_bus $timer_handle "SOPB"
#			set timer_base [xget_sw_parameter_value $timer_handle "C_BASEADDR"]
#			set timer_end [xget_sw_parameter_value $timer_handle "C_HIGHADDR"]
#			set timer_base [format "0x%08x" $timer_base]
			puts $config_file "/* Timer pheriphery is $timer */"
			puts $config_file "#define XILINX_TIMER_BASEADDR\t[uboot_addr_hex $timer_handle "C_BASEADDR"]"

#			puts "$timer_handle $intc $intc_value $intc_signals Interrupt"
			set intr [get_intr $timer_handle $intc_handle $intc_value "Interrupt"]
			puts $config_file "#define XILINX_TIMER_IRQ\t$intr"
		}
		puts $config_file ""
	}
# ******************************************************************************
# UartLite driver - I suppose, only uartlite driver
	set uart [xget_sw_parameter_value $os_handle "stdout"]
	if {[string match "" $uart] || [string match -nocase "none" $uart]} {
		puts $config_file "/* Uart not defined */"
		puts "ERROR Uart not specified. Please specific console"
	} else {
		set uart_handle [xget_sw_ipinst_handle_from_processor $proc_handle $uart]
		puts $config_file "/* Uart pheriphery is $uart */"
		set type [xget_value $uart_handle "VALUE"]
		switch $type {
			"opb_uart16550" {
				puts $config_file "#define XILINX_UART16550"
				puts $config_file "#define XILINX_UART16550_BASEADDR\t[uboot_addr_hex $uart_handle "C_BASEADDR"]"
#				FIXME bad sysclock frequency
#				puts $config_file "#define XILINX_UART16550_CLOCK_HZ\t$system_bus"
# it is driver dependent code
				set drv_handle [xget_handle $proc_handle "DRIVER" "uartns550"]
				if { [llength $drv_handle] == 0 } {
					error "Please specify your standalone driver for Uart16550"
				}
				puts $config_file "#define XILINX_UART16550_CLOCK_HZ\t[xget_value $drv_handle "PARAMETER" "CLOCK_HZ"]"
# dependent end
			}
			"opb_uartlite" -
			"xps_uartlite" {
				set args [xget_sw_parameter_handle $uart_handle "*"]
				foreach arg $args {
					set arg_name [xget_value $arg "NAME"]
					set arg_value [xget_value $arg "VALUE"]
					set arg_name [string map -nocase {C_ ""} $arg_name]
					case $arg_name in {
						"BASEADDR" {
							puts $config_file "#define XILINX_UARTLITE_${arg_name}\t$arg_value"
							set uart_base $arg_value
						}
						"BAUDRATE" {
							puts $config_file "#define XILINX_UARTLITE_${arg_name}\t$arg_value"
						}
						"HIGHADDR" {
							set uart_end $arg_value
						}
						default	{}
					}
				}
			}
			default {
				error "Unsupported type of console - $type"
			}
		}
	}
	puts $config_file ""
	# ******************************************************************************
	# IIC driver - I suppose, only uartlite driver
	set iic [xget_sw_parameter_value $os_handle "iic"]
	if {[string match "" $iic] || [string match -nocase "none" $iic]} {
		puts $config_file "/* IIC doesn't exist */"
	} else {
		set iic_handle [xget_sw_ipinst_handle_from_processor $proc_handle $iic]
		set iic_baseaddr [xget_sw_parameter_value $iic_handle "C_BASEADDR"]
		set iic_baseaddr [format "0x%08x" $iic_baseaddr]
		set iic_freq [xget_sw_parameter_value $iic_handle "C_IIC_FREQ"]
		set iic_bit [xget_sw_parameter_value $iic_handle "C_TEN_BIT_ADR"]
		puts $config_file "/* IIC pheriphery is $iic */"
		puts $config_file "#define XILINX_IIC_0_BASEADDR\t$iic_baseaddr"
		puts $config_file "#define XILINX_IIC_0_FREQ\t$iic_freq"
		puts $config_file "#define XILINX_IIC_0_BIT\t$iic_bit"
	}
	puts $config_file ""
	# ******************************************************************************
	# GPIO configuration
	set gpio [xget_sw_parameter_value $os_handle "gpio"]
	if {[string match "" $gpio] || [string match "none" $gpio]} {
		puts $config_file "/* GPIO doesn't exist */"
	} else {
		set base_param_name [format "C_BASEADDR" $gpio]
		set gpio_handle [xget_sw_ipinst_handle_from_processor $proc_handle $gpio]
		set gpio_base [xget_sw_parameter_value $gpio_handle $base_param_name]
		set gpio_base [format "0x%08x" $gpio_base]

		set gpio_end [xget_sw_parameter_value $gpio_handle "C_HIGHADDR"]
		set gpio_end [format "0x%08x" $gpio_end]

		puts $config_file "/* GPIO is $gpio*/"
		puts $config_file "#define XILINX_GPIO_BASEADDR\t$gpio_base"
	}
	puts $config_file ""
	# ******************************************************************************
	# System memory
	set main_mem [xget_sw_parameter_value $os_handle "main_memory"]
	if {[string match "" $main_mem] || [string match "none" $main_mem]} {
		puts "ERROR main_memory not specified. Please specific main_memory"
		puts $config_file "/* Main Memory doesn't exist */"
	} else {
		set main_mem_bank [xget_sw_parameter_value $os_handle "main_memory_bank"]
		set main_mem_handle [xget_sw_ipinst_handle_from_processor $proc_handle $main_mem]
		if {[string compare -nocase $main_mem_handle ""] != 0} {
			if {[xget_hw_value $main_mem_handle] == "mpmc"} {
				set base_param_name "C_MPMC_BASEADDR"
				set high_param_name "C_MPMC_HIGHADDR"
			} else {
				set base_param_name [format "C_MEM%i_BASEADDR" $main_mem_bank]
				set high_param_name [format "C_MEM%i_HIGHADDR" $main_mem_bank]
			}
			set eram_base [xget_sw_parameter_value $main_mem_handle $base_param_name]
			set eram_end [xget_sw_parameter_value $main_mem_handle $high_param_name]
			set eram_size [expr $eram_end - $eram_base + 1]
			set eram_base [format "0x%08x" $eram_base]
			set eram_size [format "0x%08x" $eram_size]
			set eram_high [expr $eram_base + $eram_size]
			set eram_high [format "0x%08x" $eram_high]
			puts $config_file "/* Main Memory is $main_mem */"
			puts $config_file "#define XILINX_RAM_START\t$eram_base"
			puts $config_file "#define XILINX_RAM_SIZE\t$eram_size"
		}
	}
	puts $config_file ""
	# ******************************************************************************
	# Flash memory
	set flash_mem [xget_sw_parameter_value $os_handle "flash_memory"]
	if {[string match "" $flash_mem] || [string match "none" $flash_mem]} {
		puts $config_file "/* FLASH doesn't exist $flash_mem */"
		puts "FLASH doesn't exists"
	} else {
		set flash_mem_handle [xget_sw_ipinst_handle_from_processor $proc_handle $flash_mem]
		set flash_mem_bank [xget_sw_parameter_value $os_handle "flash_memory_bank"]
		set base_param_name [format "C_MEM%i_BASEADDR" $flash_mem_bank]
		set high_param_name [format "C_MEM%i_HIGHADDR" $flash_mem_bank]
		set flash_start [xget_sw_parameter_value $flash_mem_handle $base_param_name]
		set flash_end [xget_sw_parameter_value $flash_mem_handle $high_param_name]
		set flash_size [expr $flash_end - $flash_start + 1]
		set flash_start [format "0x%08x" $flash_start]
		set flash_size [format "0x%08x" $flash_size]
		if {$eram_base < $flash_start} {
			puts $config_file "/* Flash Memory is $flash_mem */"
			puts $config_file "#define XILINX_FLASH_START\t$flash_start"
			puts $config_file "#define XILINX_FLASH_SIZE\t$flash_size"
		} else {
			error "Flash base address must be on higher address than ram memory"
		}
	}
	puts $config_file ""
	# ******************************************************************************
	# Sysace
	set sysace [xget_sw_parameter_value $os_handle "sysace"]
	if {[string match "" $sysace] || [string match "none" $sysace]} {
		puts $config_file "/* Sysace doesn't exist */"
	} else {
		puts $config_file "/* Sysace Controller is $sysace */"
		set sysace_handle [xget_sw_ipinst_handle_from_processor $proc_handle $sysace]
		set args [xget_sw_parameter_handle $sysace_handle "*"]
		foreach arg $args {
			set arg_name [xget_value $arg "NAME"]
			set arg_name [string map -nocase {C_ ""} $arg_name]
			set arg_value [xget_value $arg "VALUE"]
			switch $arg_name {
				"BASEADDR" {
					puts $config_file "#define XILINX_SYSACE_${arg_name}\t$arg_value"
					set sysace_base $arg_value
				}
				"MEM_WIDTH" {
					puts $config_file "#define XILINX_SYSACE_${arg_name}\t$arg_value"
				}
				"HIGHADDR" {
					set sysace_end $arg_value
				}
				"HW_VER" {
				}
				default {}
			}
		}
	}
	puts $config_file ""
	# ******************************************************************************
	# Ethernet
	set ethernet [xget_sw_parameter_value $os_handle "ethernet"]
	if {[string match "" $ethernet] || [string match -nocase "none" $ethernet]} {
		puts $config_file "/* Ethernet doesn't exist */"
	} else {
		set ethernet_handle [xget_sw_ipinst_handle_from_processor $proc_handle $ethernet]
		set ethernet_name [xget_value $ethernet_handle "VALUE"]
		puts $config_file "/* Ethernet controller is $ethernet */"

		switch $ethernet_name {
			"opb_ethernet" -
			"xps_ethernet" {
				set args [xget_sw_parameter_handle $ethernet_handle "*"]
				foreach arg $args {
					set arg_name [xget_value $arg "NAME"]
					set arg_value [xget_value $arg "VALUE"]
					set arg_name [string map -nocase {C_ ""} $arg_name]
#						{"MII_EXIST" "DMA_PRESENT" "HALF_DUPLEX_EXIST"} {
#							puts $config_file "#define XILINX_EMAC_${arg_name}\t$arg_value"
#						}
					case $arg_name in {
						"BASEADDR" {
							puts $config_file "#define XILINX_EMAC_${arg_name}\t$arg_value"
							set ethernet_base $arg_value
						}
						"HIGHADDR" {
							set ethernet_end $arg_value
						}
						default {}
					}
				}
			}
			"xps_ll_temac" {
				set args [xget_sw_parameter_handle $ethernet_handle "*"]
				foreach arg $args {
					set arg_name [xget_value $arg "NAME"]
					set arg_value [xget_value $arg "VALUE"]
					set arg_name [string map -nocase {C_ ""} $arg_name]
					#DEBUG	puts $config_file "$arg_name $arg_value"
					case $arg_name in {
						{"MII_EXIST" "DMA_PRESENT" "HALF_DUPLEX_EXIST"} {
							puts $config_file "#define XILINX_LLEMAC_${arg_name}\t$arg_value"
						}
						"BASEADDR" {
							puts $config_file "#define XILINX_LLEMAC_${arg_name}\t$arg_value"
							set ethernet_base $arg_value
						}
						"HIGHADDR" {
							set ethernet_end $arg_value
						}
						default {}
					}
				}
			}
			"opb_ethernetlite" -
			"xps_ethernetlite" {
				set args [xget_sw_parameter_handle $ethernet_handle "*"]
				foreach arg $args {
					set arg_name [xget_value $arg "NAME"]
					set arg_value [xget_value $arg "VALUE"]
					set arg_name [string map -nocase {C_ ""} $arg_name]
					case $arg_name in {
						{"BASEADDR" "TX_PING_PONG" "RX_PING_PONG"} {
							puts $config_file "#define XILINX_EMACLITE_${arg_name}\t$arg_value"
						}
						default {}
					}
				}
			}
			default {
				error "Unsupported ethernet periphery - $ethernet_name"
			}
		}
	}
	#*******************************************************************************
	# U-BOOT position in memory
# FIXME I think that generation via setting don't work corectly
	set text_base [xget_sw_parameter_value $os_handle "uboot_position"]
	set text_base [format "0x%08x" $text_base]
	puts $config_file2 ""
	if {$text_base == 0} {
		if {[llength $eram_base] != 0 } {
			puts $config_file2 "TEXT_BASE = $eram_base"
			puts "WARNING automatic U-BOOT position"
		} else {
			error "Main memory is not defined"
		}
	} elseif {![catch {unset $eram_base}]} {
		if {$eram_base < $text_base && $eram_high > $text_base} {
			#			puts $config_file2 "# TEXT BASE "
			puts $config_file2 "TEXT_BASE = $text_base"
			puts $config_file2 ""
			# print system clock
			#	set sw [xget_sw_parameter_value $proc_handle "HW_INSTANCE"]
			# FIXME Parameters for Microblaze from MHS files
			# look at Microblaze SW manual
			# print microblaze params
			#	set hwproc_handle [xget_handle $proc_handle "IPINST"]
			#	set args [xget_hw_parameter_handle $hwproc_handle "*"]
			#	set proctype [xget_value $hwproc_handle "OPTION" "IPNAME"]
		} else {
			error "ERROR u-boot position is out of range $eram_base - $eram_high"
		}
	}
}

proc headerm {ufile} {
	variable version
	puts $ufile "\#"
	puts $ufile "\# (C) Copyright 2007 Michal Simek"
	puts $ufile "\#"
	puts $ufile "\# Michal SIMEK <monstr@monstr.eu>"
	puts $ufile "\#"
	puts $ufile "\# This program is free software; you can redistribute it and/or"
	puts $ufile "\# modify it under the terms of the GNU General Public License as"
	puts $ufile "\# published by the Free Software Foundation; either version 2 of"
	puts $ufile "\# the License, or (at your option) any later version."
	puts $ufile "\#"
	puts $ufile "\# This program is distributed in the hope that it will be useful,"
	puts $ufile "\# but WITHOUT ANY WARRANTY; without even the implied warranty of"
	puts $ufile "\# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the"
	puts $ufile "\# GNU General Public License for more details."
	puts $ufile "\#"
	puts $ufile "\# You should have received a copy of the GNU General Public License"
	puts $ufile "\# along with this program; if not, write to the Free Software"
	puts $ufile "\# Foundation, Inc., 59 Temple Place, Suite 330, Boston,"
	puts $ufile "\# MA 02111-1307 USA"
	puts $ufile "\#"
	puts $ufile "\# CAUTION: This file is automatically generated by libgen."
	puts $ufile "\# Version: [xget_swverandbld]"
	puts $ufile "\# Generate by $version"
	puts $ufile "\#"
	puts $ufile ""
}

proc headerc {ufile} {
	variable version
	puts $ufile "/*"
	puts $ufile " * (C) Copyright 2007 Michal Simek"
	puts $ufile " *"
	puts $ufile " * Michal SIMEK <monstr@monstr.eu>"
	puts $ufile " *"
	puts $ufile " * This program is free software; you can redistribute it and/or"
	puts $ufile " * modify it under the terms of the GNU General Public License as"
	puts $ufile " * published by the Free Software Foundation; either version 2 of"
	puts $ufile " * the License, or (at your option) any later version."
	puts $ufile " *"
	puts $ufile " * This program is distributed in the hope that it will be useful,"
	puts $ufile " * but WITHOUT ANY WARRANTY; without even the implied warranty of"
	puts $ufile " * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the"
	puts $ufile " * GNU General Public License for more details."
	puts $ufile " *"
	puts $ufile " * You should have received a copy of the GNU General Public License"
	puts $ufile " * along with this program; if not, write to the Free Software"
	puts $ufile " * Foundation, Inc., 59 Temple Place, Suite 330, Boston,"
	puts $ufile " * MA 02111-1307 USA"
	puts $ufile " *"
	puts $ufile " * CAUTION: This file is automatically generated by libgen."
	puts $ufile " * Version: [xget_swverandbld]"
	puts $ufile " * Generate by $version"
	puts $ufile " */"
	puts $ufile ""
}

# from xillib_hw.tcl
#
# Return the handle to the bus interface associated with this address parameter.
# Parameters: periph_handle: peripheral handle
#	addr_name: name of an address parameter
#
proc xget_addrparam_busif {periph_handle addr_param_name} {
	set baseaddr_handle [xget_handle $periph_handle "PARAMETER" $addr_param_name]
	# test on null parameter
	if { [llength $baseaddr_handle] == 0 } {
		return ""
	}

	set baseaddr_bushandle [xget_handle $baseaddr_handle "SUBPROPERTY" "BUS"]
	puts "DEBUG $baseaddr_bushandle =="
	if { [llength $baseaddr_bushandle] == 0 } {
		# See if this address has a pair
		if {[string match *_BASEADDR $addr_param_name]} {
			set addr_pair [string range $addr_param_name 0 [expr [string last "_BASEADDR" $addr_param_name] -1]]
			append addr_pair "_HIGHADDR"
		} elseif {[string match *_HIGHADDR $addr_param_name]} {
			set addr_pair [string range $addr_param_name 0 [expr [string last "_HIGHADDR" $addr_param_name] -1]]
			append addr_pair "_BASEADDR"
		} else {
			return ""
		}
		set baseaddr_handle [xget_handle $periph_handle "PARAMETER" $addr_pair]
		set baseaddr_bushandle [xget_handle $baseaddr_handle "SUBPROPERTY" "BUS"]
		if { [llength $baseaddr_bushandle] == 0 } {
			return ""
		}
	}
	set baseaddr_busname [xget_value $baseaddr_handle "SUBPROPERTY" "BUS"]
	#	set baseaddr_busname [xget_value $baseaddr_bushandle "VALUE"]
	set busif_handle [xget_handle $periph_handle "BUS_INTERFACE" $baseaddr_busname]
	return $busif_handle
}

#test for peripheral - if is correct setting system bus
proc test_buses {system_bus handle bus_type} {
	set bus [xget_handle $handle "BUS_INTERFACE" $bus_type]
	if { [llength $bus] == 0 } {
		return 1
	}
	set bus [xget_value $bus "VALUE"]
	if { $bus != $system_bus} {
		error "Periphery $handle is connected to another system bus $bus ----"
		return 0
	} else {
		set name [xget_value $handle "NAME"]
		puts "$name has correct system_bus $system_bus"
	}
	return 1
}

proc xdefine_include_file2 {drv_handle file_name drv_string args} {
	foreach arg $args {
		set value [xget_value $drv_handle "PARAMETER" $arg]
		if {[llength $value] == 0} {
			error "Zero size of PARAMETER. Please append argument args."
		} else {
			return [xget_value $drv_handle "PARAMETER" $arg]
		}
	}
}

proc get_intc_signals {intc} {
	set signals [split [xget_hw_port_value $intc "intr"] "&"]
	set intc_signals {}
	foreach signal $signals {
		lappend intc_signals [string trim $signal]
	}
	return $intc_signals
}

# Get interrupt number
proc get_intr {per_handle intc intc_value port_name} {
	if {![string match "" $intc] && ![string match -nocase "none" $intc]} {
		set intc_signals [get_intc_signals $intc]
		set port_handle [xget_hw_port_handle $per_handle "$port_name"]
		set interrupt_signal [xget_value $port_handle "VALUE"]
		set index [lsearch $intc_signals $interrupt_signal]
		if {$index == -1} {
			return -1
		} else {
			# interrupt 0 is last in list.
			return [expr [llength $intc_signals] - $index - 1]
		}
	} else {
		return -1
	}
}

proc slaveip_intr {slave intc interrupt_port_list devicetype params {baseaddr_prefix ""} } {
	set tree [slaveip $slave $intc $devicetype $params $baseaddr_prefix]
	return [gen_interrupt_property $tree $slave $intc $interrupt_port_list]
}

proc slaveip {slave intc devicetype params {baseaddr_prefix ""} } {
	set baseaddr [scan_int_parameter_value $slave [format "C_%sBASEADDR" $baseaddr_prefix]]
	set highaddr [scan_int_parameter_value $slave [format "C_%sHIGHADDR" $baseaddr_prefix]]
	return [slaveip_explicit_baseaddr $slave $intc $devicetype $params $baseaddr $highaddr]
}

proc slaveip_explicit_baseaddr {slave intc devicetype params baseaddr highaddr} {
	set name [xget_hw_name $slave]
	set type [xget_hw_value $slave]
	if {$devicetype == ""} {
		set devicetype $type
	}
	set tree [slaveip_basic $slave $intc $params [format_ip_name $devicetype $baseaddr $name]]
	return [tree_append $tree [gen_reg_property $name $baseaddr $highaddr]]
}

proc slaveip_basic {slave intc params nodename} {
	set name [xget_hw_name $slave]
	set type [xget_hw_value $slave]

	set hw_ver [xget_hw_parameter_value $slave "HW_VER"]

	set ip_node {}
	lappend ip_node [gen_compatible_property $name $type $hw_ver]

	# Generate the parameters
	set ip_node [gen_params $ip_node $slave $params]

	return [list $nodename tree $ip_node]
}

proc gen_intc {slave intc devicetype param} {
	set tree [slaveip $slave $intc $devicetype $param]
	set intc_name [lindex $tree 0]
	set intc_node [lindex $tree 2]

	# Tack on the interrupt-specific tags.
	lappend intc_node [list \#interrupt-cells hexint 2]
	lappend intc_node [list interrupt-controller empty empty]
	return [list $intc_name tree $intc_node]
}

proc ll_temac_parameters {ip_handle index} {
	set params {}
	foreach param [default_parameters $ip_handle] {
		set pattern [format "C_TEMAC%d*" $index]
		if {[string match $pattern $param]} {
			lappend params $param
		}
	}
	return $params
}

proc slave_ll_temac_port {slave intc index} {
	set name [xget_hw_name $slave]
	set type [xget_hw_value $slave]
	set baseaddr [scan_int_parameter_value $slave "C_BASEADDR"]
	set baseaddr [expr $baseaddr + $index * 0x40]
	set highaddr [expr $baseaddr + 0x3f]

	set ip_tree [slaveip_basic $slave $intc "" [format_ip_name "ethernet" $baseaddr]]
	set ip_tree [tree_append $ip_tree [list "device_type" string "network"]]
	set ip_tree [tree_append $ip_tree [list "local-mac-address" bytesequence [list 00 00 00 00 00 00]]]

	set ip_tree [tree_append $ip_tree [gen_reg_property $name $baseaddr $highaddr]]
	set ip_tree [gen_interrupt_property $ip_tree $slave $intc [format "TemacIntc%d_Irpt" $index]]
	set ip_name [lindex $ip_tree 0]
	set ip_node [lindex $ip_tree 2]
	# Generate the parameters, stripping off the right prefix.
	set ip_node [gen_params $ip_node $slave [ll_temac_parameters $slave $index] [format "C_TEMAC%i_" $index]]
	# Generate the common parameters.
	set ip_node [gen_params $ip_node $slave [list "C_PHY_TYPE" "C_TEMAC_TYPE" "C_BUS2CORE_CLK_RATIO"]]
	set ip_tree [list $ip_name tree $ip_node]
	set mhs_handle [xget_hw_parent_handle $slave]
	# See what the temac is connected to.
	set ll_busif_handle [xget_hw_busif_handle $slave "LLINK$index"]
	set ll_name [xget_hw_value $ll_busif_handle]
	set ll_ip_handle [xget_hw_connected_busifs_handle $mhs_handle $ll_name "target"]
	set ll_ip_handle_name [xget_hw_name $ll_ip_handle]
	set connected_ip_handle [xget_hw_parent_handle $ll_ip_handle]
	set connected_ip_name [xget_hw_name $connected_ip_handle]
	set connected_ip_type [xget_hw_value $connected_ip_handle]
	if {$connected_ip_type == "mpmc"} {
		# Assumes only one MPMC.
		if {[string match SDMA_LL? $ll_ip_handle_name]} {
			set port_number [string range $ll_ip_handle_name 7 7]
			set sdma_name "PIM$port_number"
			set ip_tree [tree_append $ip_tree [list "llink-connected" labelref $sdma_name]]
		} else {
			error "found ll_temac connected to mpmc, but can't find the port number!"
		}
	} else {
		# Hope it's something that only has one locallink
		# connection. Most likely an xps_ll_fifo
		set ip_tree [tree_append $ip_tree [list "llink-connected" labelref "$connected_ip_name"]]
	}
	return $ip_tree
}
proc slave_ll_temac {slave intc} {
	set baseaddr [scan_int_parameter_value $slave "C_BASEADDR"]
	set ip_name [xget_hw_name $slave]
	set ip_type [xget_hw_value $slave]
	set tree [list [format_ip_name $ip_type $baseaddr $ip_name] tree {}]
	set tree [tree_append $tree [list \#size-cells hexint 1]]
	set tree [tree_append $tree [list \#address-cells hexint 1]]
	set tree [tree_append $tree [slave_ll_temac_port $slave $intc 0] ]
	set port1_enabled  [scan_int_parameter_value $slave "C_TEMAC1_ENABLED"]
	if {$port1_enabled == "1"} {
		set tree [tree_append $tree [slave_ll_temac_port $slave $intc 0] ]
	}
	return $tree
}
proc slave_mpmc {slave intc} {
	if {[catch {
		# Found control port for ECC and performance monitors
		set tree [slaveip $slave $intc "" "" "MPMC_CTRL_"]
		set ip_name [lindex $tree 0]
		set mpmc_node [lindex $tree 2]
	}]} {
		# No control port
		set baseaddr [scan_int_parameter_value $slave "C_MPMC_BASEADDR"]
		set tree [slaveip_basic $slave $intc "" [format_ip_name "mpmc" $baseaddr] ]
		set ip_name [lindex $tree 0]
		set mpmc_node [lindex $tree 2]

		# Generate the parameters
		# set mpmc_node [gen_params $mpmc_node $slave [default_parameters $slave] ]

	}
	lappend mpmc_node [list \#size-cells hexint 1]
	lappend mpmc_node [list \#address-cells hexint 1]

	set num_ports [scan_int_parameter_value $slave "C_NUM_PORTS"]
	set share_addresses [scan_int_parameter_value $slave "C_ALL_PIMS_SHARE_ADDRESSES"]
	for {set x 0} {$x < $num_ports} {incr x} {
		set pim_type [scan_int_parameter_value $slave [format "C_PIM%d_BASETYPE" $x]]
		if {$pim_type == 3} {
			# Found an SDMA port
			if {$share_addresses == 0} {
				set baseaddr [scan_int_parameter_value $slave [format "C_SDMA_CTRL%d_BASEADDR" $x]]
			} else {
				set baseaddr [scan_int_parameter_value $slave "C_SDMA_CTRL_BASEADDR"]
			}
			set baseaddr [expr $baseaddr + $x * 0x80]
			set highaddr [expr $baseaddr + 0x7f]

			set sdma_name [format_ip_name sdma $baseaddr "PIM$x"]
			set sdma_tree [list $sdma_name tree {}]
			set sdma_tree [tree_append $sdma_tree [gen_reg_property $sdma_name $baseaddr $highaddr]]
			set sdma_tree [tree_append $sdma_tree [gen_compatible_property $sdma_name "ll_dma" "1.00a"]]
			set sdma_tree [gen_interrupt_property $sdma_tree $slave $intc [list [format "SDMA%d_Rx_IntOut" $x] [format "SDMA%d_Tx_IntOut" $x]]]

			lappend mpmc_node $sdma_tree

		}
	}
	return [list $ip_name tree $mpmc_node]
}

#
#get handle to interrupt controller from CPU handle
#
proc get_handle_to_intc {proc_handle port_name} {
	#one CPU handle
	set hwproc_handle [xget_handle $proc_handle "IPINST"]
	#hangle to mhs file
	set mhs_handle [xget_hw_parent_handle $hwproc_handle]
	#get handle to interrupt port on Microblaze
	set intr_port [xget_value $hwproc_handle "PORT" $port_name]
	if { [llength $intr_port] == 0 } {
		error "CPU has not connection to Interrupt controller"
	}
	#	set sink_port [xget_hw_connected_ports_handle $mhs_handle $intr_port "sink"]
	#	set sink_name [xget_hw_name $sink_port]
	#get source port periphery handle - on interrupt controller
	set source_port [xget_hw_connected_ports_handle $mhs_handle $intr_port "source"]
	#get interrupt controller handle
	set intc [xget_hw_parent_handle $source_port]
	#	set name [xget_hw_name $intc]
	#	puts "$intc $name"
	return $intc
}

# Populates a bus node with components connected to the given slave
# and adds it to the given tree
#
# tree         : Tree to populate
# slave_handle : The slave to use as a starting point, this is
# typically the root processor or a previously traversed bus bridge.
# intc_handle	: The interrupt controller associated with the
# processor. Slave will have an interrupts node relative to this
# controller.
# face		   : The name of the port of the slave that is connected to the bus.
# baseaddr_prefix : prefix used to query the slave for address
# translation ranges, or empty to generate a 1:1 translation.
proc bus_bridge {tree slave intc_handle face baseaddr_prefix} {
	puts "+++++++++++ $slave ++++++++"
	set busif_handle [xget_hw_busif_handle $slave $face]
	if {[llength $busif_handle] == 0} {
		return $tree
	}
	set bus_name [xget_hw_value $busif_handle]

	#	puts "--$busif_handle"
	set mhs_handle [xget_hw_parent_handle $slave]
	set bus_handle [xget_hw_ipinst_handle $mhs_handle $bus_name]

	if {[llength $bus_handle] == 0} {
		return $tree
	}
	set bus_type [xget_hw_value $bus_handle]
	set hw_ver [xget_hw_parameter_value $bus_handle "HW_VER"]

	set master_ifs [xget_hw_connected_busifs_handle $mhs_handle $bus_name "master"]
	set slave_ifs [xget_hw_connected_busifs_handle $mhs_handle $bus_name "slave"]
	set bus_ip_handles {}
	# Compose peripherals & cleaning
	foreach if $master_ifs {
		set ip_handle [xget_hw_parent_handle $if]
		puts "-master $ip_handle [xget_hw_name $if] [xget_hw_value $if] [xget_hw_name $ip_handle]"

		# If its not already in the list, and its not the bridge, then
		# append it.
		if {$ip_handle != $slave} {
			if {[lsearch $bus_ip_handles $ip_handle] == -1} {
				lappend bus_ip_handles $ip_handle
			}
		}
	}

	foreach if $slave_ifs {
		set ip_handle [xget_hw_parent_handle $if]
		puts "-slave $ip_handle [xget_hw_name $if] [xget_hw_value $if] [xget_hw_name $ip_handle]"

		# If its not already in the list, and its not the bridge, then
		# append it.
		if {$ip_handle != $slave} {
			if {[lsearch $bus_ip_handles $ip_handle] == -1} {
				lappend bus_ip_handles $ip_handle
			}
		}
	}
	variable periphery_array

	# Start generating the node for the bus.
	set bus_node {}

	# Generate the ranges
	# 	if {$baseaddr_prefix != ""} {
	# 		set count [scan_int_parameter_value $slave "C_NUM_ADDR_RNG"]
	# 		if { [llength $count] == 0 } {
	# 			set count 1
	# 		}
	# 		for {set x 0} {$x < $count} {incr x} {
	# 			set baseaddr [scan_int_parameter_value $slave [format "%s%s_BASEADDR" $baseaddr_prefix $x]]
	# 			set highaddr [scan_int_parameter_value $slave [format "%s%s_HIGHADDR" $baseaddr_prefix $x]]

	# 			set bus_node [gen_ranges_property $bus_node $bus_name $x $baseaddr $highaddr]
	# 		}
	# 	 } else {
	lappend bus_node [list ranges empty empty]
	# 	 }

	# Populate with all the slaves.
	foreach ip $bus_ip_handles {
		# If we haven't already generated this ip
		if {[lsearch $periphery_array $ip] == -1} {
			set bus_node [gener_slave $bus_node $ip $intc_handle]
			lappend periphery_array $ip
		}
	}

	lappend bus_node [list \#size-cells hexint 1]
	lappend bus_node [list \#address-cells hexint 1]
	lappend bus_node [gen_compatible_property $bus_name $bus_type $hw_ver]

	lappend tree [list $bus_type tree $bus_node]
	return $tree
}

# Return a sorted list of all the port names that we think are
# interrupts (i.e. those tagged in the mpd with SIGIS=INTERRUPT)
proc interrupt_list {ip_handle} {
	set port_handles [xget_hw_port_handle $ip_handle "*"]
	set interrupt_ports {}
	foreach port $port_handles {
		set name [xget_value $port "NAME"]
		set list [xget_hw_subproperty_handle $port "SIGIS"]
		set sigis [xget_hw_subproperty_value $port "SIGIS"]
		if {[string match $sigis "INTERRUPT"]} {
			lappend interrupt_ports $name
		}
	}
	return [lsort $interrupt_ports]
}

# Return a list of all the parameter names for the given ip that
# should be reported in the device tree for generic IP. This list
# includes all the parameter names, except those that are handled
# specially, such as the instance name, baseaddr, etc.
proc default_parameters {ip_handle} {
	set par_handles [xget_hw_parameter_handle $ip_handle "*"]
	set params {}
	foreach par $par_handles {
		set par_name [xget_hw_name $par]
		# Ignore some parameters that are always handled specially
		switch -glob $par_name {
			"INSTANCE" -
			"C_FAMILY" -
			"*BASEADDR" -
			"*HIGHADDR" -
			"C_SPLB*" -
			"C_OPB*" -
			"C_DPLB*" -
			"C_IPLB*" -
			"C_PLB*" -
			"HW_VER" {}
			default { lappend params $par_name }
		}
	}
	return $params
}

proc scan_int_parameter_value {ip_handle name} {
	set param_handle [xget_hw_parameter_handle $ip_handle $name]
	if {$param_handle == ""} {
		error "Can't find parameter $name in [xget_hw_name $ip_handle]"
		return 0
	}
	set value [xget_hw_value $param_handle]
	# tcl 8.4 doesn't handle binary literals..
	if {[string match 0b* $value]} {
		# Chop off the 0b
		set tail [string range $value 2 [expr [string length $value]-1]]
		# Convert the remainder back to decimal
		binary scan [binary format "B*" $tail] "I*" value
	}
	return [expr $value]
}

proc format_name {par_name} {
	set par_name [string tolower $par_name]
	set par_name [string map -nocase {"_" "-"} $par_name]
	return $par_name
}

proc format_xilinx_name {name} {
	return "xlnx,[format_name $name]"
}

proc format_param_name {name trimprefix} {
	if {[string match [string range $name 0 [expr [string length $trimprefix] - 1]] $trimprefix]} {
		set name [string range $name [string length $trimprefix] [string length $name]]
	}
	return [format_xilinx_name $name]
}

proc format_ip_name {devicetype baseaddr {label ""}} {
	set node_name [format_name [format "%s@%x" $devicetype $baseaddr]]
	if {[string match $label ""]} {
		return $node_name
	} else {
		return [format "%s: %s" $label $node_name]
	}
}

proc gen_params {node_list handle params {trimprefix "C_"} } {
	foreach par_name $params {
		if {[catch {
			set par_value [scan_int_parameter_value $handle $par_name]
			lappend node_list [list [format_param_name $par_name $trimprefix] hexint $par_value]
		} {err}]} {
			set par_handle [xget_hw_parameter_handle $handle $par_name]
			if {$par_handle == ""} {
				puts "Unknown parameter name $i"
			} else {
				set par_value [xget_hw_value $par_handle]
			}
			lappend node_list [list [format_param_name $par_name $trimprefix] string $par_value]
		}
	}
	return $node_list
}

proc gen_compatible_property {nodename type hw_ver} {
	set namewithver [format "%s_%s" $type $hw_ver]
	return [list "compatible" stringtuple [list [format_xilinx_name "$namewithver"] [format_xilinx_name "$type"]]]
}

proc gen_ranges_property {node nodename index baseaddr highaddr} {
	if { ![llength $baseaddr] || ![llength $highaddr] } {
		error "Bad address range $nodename"
	}
	if {[string match $highaddr "0x00000000"]} {
		return $node
	}
	set size [expr $highaddr - $baseaddr + 1]
	if { $size < 0 } {
		error "Bad highaddr for $nodename"
	}
	lappend node [list "ranges" hexinttuple [list $index $baseaddr $size] ]
	return $node
}

proc gen_interrupt_property {tree slave intc interrupt_port_list} {
	set pocet [scan_int_parameter_value $intc "C_NUM_INTR_INPUTS"]
	set pocet [expr $pocet - 1]
	set intc_name [xget_hw_name $intc]
	set interrupt_list {}
	foreach in $interrupt_port_list {
		set irq [get_intr $slave $intc $pocet $in]
		if {![string match $irq "-1"]} {
			lappend interrupt_list "$irq" 0
		}
	}
	if {[llength $interrupt_list] != 0} {
		set tree [tree_append $tree [list "interrupts" hexinttuple $interrupt_list]]
		set tree [tree_append $tree [list "interrupt-parent" labelref $intc_name]]
	}
	return $tree
}

proc gen_reg_property {nodename baseaddr highaddr} {
	if { ![llength $baseaddr] || ![llength $highaddr] } {
		error "Bad address range $nodename"
	}
	if {[string match $highaddr "0x00000000"]} {
		error "No high address for $nodename"
	}
	set size [expr $highaddr - $baseaddr + 1]
	if { $size < 0 } {
		error "Bad highaddr for $nodename"
	}
	return [list "reg" hexinttuple [list $baseaddr $size]]
}

proc write_value {file indent type value} {
	if {[catch {
		if {$type == "int"} {
			#	puts -nonewline $file "= [format %d $value]"
			puts -nonewline $file "= <[format %x $value]>"
		} elseif {$type == "hexint"} {
			puts -nonewline $file "= <[format %x $value]>"
		} elseif {$type == "empty"} {
		} elseif {$type == "hexinttuple"} {
			puts -nonewline $file "= < "
			foreach element $value {
				puts -nonewline $file "[format %x $element] "
			}
			puts -nonewline $file ">"
		} elseif {$type == "bytesequence"} {
			puts -nonewline $file "= \[ "
			foreach element $value {
				if {[expr $element > 255]} {
					error {"Value $element is not a byte!"}
				}
				puts -nonewline $file "[format %02x $element] "
			}
			puts -nonewline $file "\]"
		} elseif {$type == "labelref"} {
			puts -nonewline $file "= <&$value>"
		} elseif {$type == "string"} {
			puts -nonewline $file "= \"$value\""
		} elseif {$type == "stringtuple"} {
			puts -nonewline $file "= "
			set first true
			foreach element $value {
				if {$first != true} { puts -nonewline $file ", " }
				puts -nonewline $file "\"$element\""
				set first false
			}
		} elseif {$type == "tree"} {
			write_tree $indent $file $value
		} else {
			puts "unknown type $type"
		}
	} {error}]} {
		puts $error
		puts -nonewline $file "= \"$value\""
	}
	puts $file ";"
}

# tree: a tree triple
# child_node: a tree triple
# returns: tree with child_node appended to the list of child nodes
proc tree_append {tree child_node} {
	if {[lindex $tree 1] != "tree"} {
		error {"tree_append called on $tree, which is not a tree."}
	}
	set name [lindex $tree 0]
	set node [lindex $tree 2]
	lappend node $child_node
	return [list $name tree $node]
}

proc write_nodes {indent file tree} {
	set tree [lsort -index 0 $tree]
	foreach node $tree {
		if { [llength $node] == 3} {
			set name [lindex $node 0]
			set type [lindex $node 1]
			set value [lindex $node 2]
			puts -nonewline $file "[tt [expr $indent + 1]]$name "
			write_value $file [expr $indent + 1] $type $value
		} else {
			puts "Error_bad_tree_node length = [llength $node], $node"
		}
	}
}

proc write_tree {indent file tree} {
	puts $file "{"
	set trees {}
	set nontrees {}
	foreach node $tree {
		if { [string match [lindex $node 1] "tree"]} {
			lappend trees $node
		} else {
			lappend nontrees $node
		}
	}
	write_nodes $indent $file $nontrees
	write_nodes $indent $file $trees

	puts -nonewline $file "[tt $indent]} "
}

# Generate an integer-typed parameter into config.in and C header format
proc generate_int_param {param_name param_value descr} {
    global kconfig_file

    puts $kconfig_file "config ${param_name}"
    puts $kconfig_file "       int \"${descr}\" if ALLOW_EDIT_AUTO"
    puts $kconfig_file "       default ${param_value}"
    puts $kconfig_file ""
}

# Generate a string-typed parameter into config.in and C header format
proc generate_string_param {param_name param_value descr} {
    global kconfig_file

    puts $kconfig_file "config ${param_name}"
    puts $kconfig_file "       string \"${descr}\" if ALLOW_EDIT_AUTO"
    puts $kconfig_file "       default ${param_value}"
    puts $kconfig_file ""
}

# Generate a comment into config.in and C header format
proc generate_comment {comment_string} {
    global kconfig_file
    puts $kconfig_file "\# $comment_string"
    puts $kconfig_file "comment \"${comment_string}\""
    puts $kconfig_file "      depends on ALLOW_EDIT_AUTO"
    puts $kconfig_file ""
}

# Generate a hex-typed parameter into config.in and C header format
proc generate_hex_param {param_name param_value descr} {
    global kconfig_file

    puts $kconfig_file "config ${param_name}"
    puts $kconfig_file "       hex \"${descr}\""
    puts $kconfig_file "       default ${param_value}"
    puts $kconfig_file ""
}

